from unittest.mock import MagicMock

import pytest

from fgn.utils.output_manager import OutputManager


# Set up your mocks at the top of the file
@pytest.fixture(autouse=True)
def mock_open(mocker):
    mocker.patch("builtins.open", mocker.mock_open())
    return open


@pytest.fixture(autouse=True)
def mock_pyperclip_copy(mocker):
    mock_copy = mocker.patch("pyperclip.copy")
    return mock_copy


@pytest.fixture
def output_manager():
    return OutputManager()


def test_handle_output_no_flags(mock_open, mock_pyperclip_copy, output_manager):
    response = "Test response"
    output_manager.handle_output(response)

    mock_pyperclip_copy.assert_called_once_with(response)
    mock_open.assert_not_called()  # since neither 'output' nor 'auto_output' flags are set


def test_handle_output_output_flag(mock_open, mock_pyperclip_copy, output_manager):
    response = "Test response"
    output_manager.output = "test_output.md"
    output_manager.handle_output(response)

    mock_pyperclip_copy.assert_called_once_with(response)
    mock_open.assert_called_once()  # since 'output' flag is set


def test_handle_output_auto_output_flag(mock_open, mock_pyperclip_copy, output_manager):
    response = "Test response"
    output_manager.auto_output = True
    output_manager.handle_output(response)

    mock_pyperclip_copy.assert_called_once_with(response)
    mock_open.assert_called()  # since 'auto_output' flag is set


def test_handle_output_no_copy_flag(mock_open, mock_pyperclip_copy, output_manager):
    response = "Test response"
    output_manager.no_copy = True
    output_manager.handle_output(response)

    mock_pyperclip_copy.assert_not_called()  # since 'no_copy' flag is set
    mock_open.assert_not_called()  # since neither 'output' nor 'auto_output' flags are set


def test_save_to_file_no_filename(mock_open, output_manager):
    response = "Test response"
    output_manager.save_to_file(response)

    mock_open.assert_called()
    handle = mock_open()
    handle.write.assert_called_with(response)


def test_save_to_file_with_filename(mock_open, output_manager):
    response = "Test response"
    filename = "test_output.md"
    output_manager.save_to_file(response, filename)

    mock_open.assert_called_once_with(filename, "w")
    file_handle = mock_open.return_value.__enter__.return_value
    file_handle.write.assert_called_with(response)


def test_save_to_file_append(mock_open, output_manager):
    response = "Test response"
    filename = "test_output.md"
    output_manager.save_to_file(response, filename, append=True)

    mock_open.assert_called_once_with(filename, "a")
    file_handle = mock_open.return_value.__enter__.return_value
    file_handle.write.assert_called_with("\n\n" + response)



import pyperclip


def paste_into_fgn():
    # print(f"Pasting into fgn {pyperclip.paste()}")
    return pyperclip.paste()


def copy_into_clipboard(text):
    # print(f"Copying into clipboard {text}")
    pyperclip.copy(text)



# fgn/core/commands/core/core_cmd.py
from pathlib import Path

import click

from fgn.core.core_cmd import core_command
from fgn.core.default_sub_cmd import default_sub_cmd

cmd_name = Path(__file__).stem.split("_")[0]


@click.command()
@click.pass_context
def cli(ctx: click.Context) -> None:
    """CLI internal usage only. Do not call directly."""
    core_command(ctx.obj)



from dataclasses import dataclass, field
from typing import List, Optional

from fgn.models.dsl.task import Task


@dataclass
class FgnDslSchema:
    version: Optional[str] = None
    description: Optional[str] = None
    tasks: List[Task] = field(default_factory=list)



from dataclasses import dataclass, field
from typing import List, Optional

from fgn.models.dsl.step import Step


@dataclass
class Task:
    name: str
    description: Optional[str] = None
    steps: List[Step] = field(default_factory=list)



from pathlib import Path

import click

from fgn.utils.file_operations import create_project_dir


@click.command()
@click.argument("message", default="Hello, World", required=False)
@click.option("-o", "--output", help="Output file to write the message.")
def main(message, output):
    create_project_dir()

    click.echo(message)

    # If the output option is provided, write the message to the file
    if output:
        with open(output, "w") as file:
            file.write(message)


if __name__ == "__main__":
    main()



import pytest
from pyfakefs.fake_filesystem_unittest import Patcher

from fgn.core.command_context import CommandContext
from fgn.utils.file_operations import (
    extract_markdown,
    get_norm_path,
    get_project_root,
    load_default_or_context,
    open_file,
    open_file_or_raise,
    save_relative_to_base,
)


@pytest.fixture
def fs():
    patcher = Patcher()
    patcher.setUp()
    yield patcher.fs
    patcher.tearDown()


def test_load_default_or_context():
    # Here, you would write tests for the load_default_or_context function.
    # You would need to create a variety of test cases to cover all possibilities.
    pass


def test_open_file(fs):
    fs.create_file("/test.txt", contents="test content")
    assert open_file("/test.txt") == "test content"


def test_save_relative_to_base(fs):
    base_path = "/path/to/base"
    fs.create_dir(base_path)
    save_relative_to_base("test.txt", "test content", base_path)
    with open("/path/to/base/test.txt") as f:
        assert f.read() == "test content"


def test_open_file_or_raise(fs):
    with pytest.raises(FileNotFoundError):
        open_file_or_raise("/nonexistent.txt")
    fs.create_file("/test.txt", contents="test content")
    assert open_file_or_raise("/test.txt") == "test content"


def test_extract_markdown():
    assert (
        extract_markdown("```test\nprint('Hello World!')```") == "print('Hello World!')"
    )
    assert extract_markdown("no markdown here") == "no markdown here"


def test_get_project_root():
    assert str(get_project_root()).endswith("fgn")


def test_get_norm_path_windows(mocker):
    mocker.patch("os.name", "nt")
    assert get_norm_path("/path/to/file") == "\\path\\to\\file"


def test_get_norm_path_non_windows(mocker):
    mocker.patch("os.name", "posix")
    assert get_norm_path("/path/to/file") == "/path/to/file"



# fgn/utils/output_manager.py
from dataclasses import dataclass
from typing import Optional

import pyperclip
from rich import print
from rich.markdown import Markdown

from fgn.utils.file_operations import extract_markdown
from fgn.utils.openai_operations import generate_output_file


@dataclass
class OutputManager:
    output: Optional[str] = None
    no_copy: bool = False
    auto_output: bool = False
    verbose: bool = False
    extension: str = "md"

    def handle_output(
        self, response: str, extract_md: bool = False, append: Optional[str] = False
    ) -> None:
        if extract_md:
            response = extract_markdown(response)
            if self.verbose:
                print("Extracting markdown...")
                print(response)
        output = self.output
        no_copy = self.no_copy
        auto_output = self.auto_output

        if output:
            self.save_to_file(response, output, append=append)
        if auto_output:
            self.save_to_file(response, extension=self.extension)
        if not no_copy:
            pyperclip.copy(response)
            if self.verbose:
                print("The output has been saved to clipboard.")

        md = Markdown(str(response))
        print(md)

    def save_to_file(
        self,
        response: str,
        filename: Optional[str] = None,
        extension: str = "md",
        append: Optional[bool] = False,
    ) -> None:
        if not filename:
            filename = generate_output_file(response, extension=extension)
        if append:
            with open(filename, "a") as output_file:
                output_file.write("\n\n" + response)
        else:
            with open(filename, "w") as output_file:
                output_file.write(response)
        if self.verbose:
            print(f"The output has been saved to {filename}.")



import click
from rich import print

from fgn.utils.chat_agent_configuration import ChatAgentConfiguration
from fgn.utils.output_manager import OutputManager


def build_chat_agent_and_prompt(config):
    chat_prompt = config.build_chat_prompt()
    chat_agent = config.get_chat_agent()
    return chat_agent, chat_prompt


def handle_response(config, response, extract_md, append, extension):
    output_manager = OutputManager(
        output=config.get("output"),
        no_copy=config.get("no_copy"),
        auto_output=config.get("auto_output"),
        verbose=config.get("verbose"),
        extension=extension,
    )
    output_manager.handle_output(response, extract_md=extract_md, append=append)


def default_sub_cmd(
    ctx, cmd_name: str, extract_md=False, prompt_prefix="", prompt_suffix=""
):
    """Execute the default sub command."""

    fgn_context = ctx.obj
    config = ChatAgentConfiguration(fgn_context, cmd_name)
    chat_agent, chat_prompt = build_chat_agent_and_prompt(config)

    if config.get("verbose"):
        print(f"Running {cmd_name} command...")

    if prompt_prefix:
        chat_prompt = f"{prompt_prefix} {chat_prompt}"

    if prompt_suffix:
        chat_prompt = f"{chat_prompt} {prompt_suffix}"

    if config.get("verbose"):
        print(f"Input: {chat_prompt}")

    response = chat_agent.submit(chat_prompt)

    handle_response(
        config, response, extract_md, fgn_context.append, fgn_context.extension
    )

    if config.get("verbose"):
        print(f"Output: {response}")



from dataclasses import dataclass


@dataclass
class Message:
    role: str
    content: str

    def serialize(self):
        return {"role": self.role, "content": self.content}

    @classmethod
    def deserialize(cls, d):
        role = d["role"]
        content = d["content"]
        return cls(role, content)

    def __str__(self):
        return f"{self.content}"



import os
import shutil
import tempfile
from unittest.mock import MagicMock, patch

import pytest

from fgn.utils.openai_operations import (
    generate_filename,
    generate_output_file,
    gpt3_completion,
    gpt4_completion,
    gpt_chat_completion,
    save_completion,
)


@pytest.fixture(autouse=True)
def mock_save_to_project_folder(monkeypatch):
    # Mock save_to_project_folder
    mock_save = MagicMock()
    monkeypatch.setattr("fgn.utils.openai_operations.save_to_project_folder", mock_save)
    return mock_save


# Use pytest fixtures for setup and teardown
@pytest.fixture
def openai_operations_fixture():
    with patch("openai.Completion.create") as mock_openai_completion, patch(
        "openai.ChatCompletion.create"
    ) as mock_openai_chat_completion:
        temp_dir = tempfile.mkdtemp()
        yield mock_openai_completion, mock_openai_chat_completion, temp_dir
        shutil.rmtree(temp_dir)


def test_gpt3_completion(openai_operations_fixture):
    mock_openai_completion, _, _ = openai_operations_fixture
    completion_text = "Test GPT-3 completion"
    mock_openai_completion.return_value = {"choices": [{"text": completion_text}]}
    result = gpt3_completion("Test prompt")
    assert result.strip() == completion_text.strip()


def test_chat_completion(openai_operations_fixture):
    _, mock_openai_chat_completion, _ = openai_operations_fixture
    completion_text = "Test chat completion"
    mock_openai_chat_completion.return_value = {
        "choices": [{"message": {"content": completion_text}}]
    }
    result = gpt_chat_completion(
        messages=[{"role": "user", "content": "Test prompt"}], model="gpt-4"
    )
    assert result.strip() == completion_text.strip()


@patch("fgn.utils.openai_operations.gpt_chat_completion")
def test_gpt4_completion(mock_chat_completion):
    completion_text = "Test GPT-4 completion"
    mock_chat_completion.return_value = completion_text
    result = gpt4_completion("Test prompt")
    assert result.strip() == completion_text.strip()


@patch("fgn.utils.openai_operations.gpt3_completion")
def test_generate_filename(mocked_gpt3_completion):
    mocked_gpt3_completion.return_value = "test_filename"
    filename = generate_filename(
        prompt="Test prompt",
        prefix="prefix",
        suffix="suffix",
        extension="py",
        max_chars=60,
        time=False,
    )
    assert filename == "prefix_test_filename_suffix.py"


@patch("fgn.utils.openai_operations.gpt3_completion")
def test_generate_output_file(mocked_gpt3_completion):
    mocked_gpt3_completion.return_value = "test_output_file"
    output_file = generate_output_file(
        prompt="Test prompt", extension="py", max_chars=60, time=False
    )
    assert output_file == "test_output_file.py"


@patch("fgn.utils.openai_operations.openai.Completion.create")
def test_gpt3_completion_with_different_stop_sequences(mock_openai_completion):
    completion_text = "Test GPT-3 completion with stop sequence"
    mock_openai_completion.return_value = {"choices": [{"text": completion_text}]}
    result = gpt3_completion("Test prompt", stop=["<<STOP>>", "<<END>>"])
    assert result.strip() == completion_text.strip()


@patch("fgn.utils.openai_operations.openai.ChatCompletion.create")
def test_gpt_chat_completion_max_context_length_error(mock_openai_chat_completion):
    error_message = "Error: Maximum context length exceeded"
    mock_openai_chat_completion.side_effect = Exception(error_message)
    result = gpt_chat_completion(
        messages=[{"role": "user", "content": "Test prompt"}],
        model="gpt-4",
        max_retry=3,
        backoff_factor=0.0001,
        initial_wait=0.0001,
    )
    assert result == f"GPT error: {error_message}"



# fgn/utils/chat_agent_configuration.py
import os
from dataclasses import dataclass
from typing import Any, Dict, Optional

import yaml

from fgn.core.chat_agent import ChatAgent
from fgn.core.command_context import CommandContext
from fgn.utils.clipboard import paste_into_fgn
from fgn.utils.file_operations import get_project_root, open_file, open_file_or_raise


@dataclass
class ChatAgentConfiguration:
    context: CommandContext
    cmd_name: str
    defaults: Dict[str, Any] = None

    def __post_init__(self):
        if self.defaults is None:
            self.defaults = self.load_default_yaml()

    def load_file_content(self, file_name: str) -> Optional[str]:
        file_path = os.path.join(
            get_project_root(), "core", "commands", self.cmd_name, file_name
        )

        if os.path.isfile(file_path):
            with open(file_path, encoding="utf-8") as file:
                return file.read()

        return None

    def get(self, key: str, default: Any = None) -> Any:
        ctx_value = getattr(self.context, key, None)
        default_value = self.defaults.get(key, default)
        return ctx_value if ctx_value is not None else default_value

    def build_chat_prompt(self) -> str:
        chat_prompt = ""
        prompt = self.get("prompt")
        # get file or throw

        paste = self.context.paste

        # Retrieve the file names from the YAML configuration or context
        example_file = self.context.example
        template_file = self.context.template
        schema_file = self.context.schema

        # Load the file content using the load_file_content() method
        example = open_file(example_file) if example_file else None
        template = open_file(template_file) if template_file else None
        schema = open_file(schema_file) if schema_file else None

        if schema:
            chat_prompt += schema + "\n"
        if template:
            chat_prompt += template + "\n"
        if example:
            chat_prompt += example + "\n"
        if self.context.input:
            chat_prompt += open_file_or_raise(self.context.input) + "\n"
        if paste:
            chat_prompt += paste_into_fgn() + "\n"
        if prompt:
            chat_prompt += prompt + " "
        if self.context.text:
            chat_prompt += self.context.text + " "

        return chat_prompt

    def load_default_yaml(self) -> Dict[str, Any]:
        # Define the new location for the YAML files
        yaml_file = os.path.join(
            get_project_root(),
            "core",
            "commands",
            self.cmd_name,
            f"{self.cmd_name}_config.yml",
        )

        config = {}
        with open(yaml_file) as file:
            config = yaml.safe_load(file)
        return config.get("defaults")

    def get_chat_agent(self) -> ChatAgent:
        chat_agent = ChatAgent(
            history_path=self.get("history_path"),
            model=self.get("model"),
            system_prompt=self.load_file_content(self.get("system_prompt")),
            auto_clear=self.get("clear_history", False),
            verbose=self.get("verbose", False),
            auto_summarize=self.get("auto_summarize", 4),
            tokens=self.context.tokens,
        )
        return chat_agent



import json
import os
from dataclasses import dataclass
from typing import List

from fgn.models.message import Message
from fgn.utils.file_operations import save_to_project_folder
from fgn.utils.openai_operations import gpt_chat_completion


@dataclass
class ChatAgent:
    model: str
    system_prompt: str = None
    messages: List[Message] = None
    auto_summarize: int = 4
    auto_clear: bool = False
    verbose: bool = False
    tokens: str = None
    history_path: str = None

    def __post_init__(self):
        if not self.messages:
            self.messages = []
        if self.system_prompt:
            self.messages.append(Message("system", self.system_prompt))
        if self.auto_clear:
            self.clear()
        else:
            if self.history_path:
                self.load()

    def submit(self, content, tokens=None):
        if self.verbose:
            print("ChatAgent.submit: ", content)

        # If tokens are provided, replace any {key} with the corresponding value
        # separated by a semicolon
        if self.tokens:
            for token in self.tokens.split(";"):
                key, value = token.split("=")
                content = content.replace("{{" + key + "}}", value)

        self.add_message("user", content)

        response = self.generate_response()

        # Check if the response contains the "maximum context length" error
        if "maximum context length" in response:
            # Auto-summarize the conversation
            success = self.summarize_conversations(self.auto_summarize)

            # If the summary was successful, generate a new response
            if success:
                response = self.generate_response()

        return response

    def add_message(self, role, content):
        self.messages.append(Message(role, content))

    def save(self):
        input_data = {"messages": [message.serialize() for message in self.messages]}

    def load(self):
        if os.path.exists(self.history_path):
            try:
                with open(self.history_path, encoding="utf-8") as infile:
                    input_data = json.load(infile)
                    self.messages = [
                        Message.deserialize(message_data)
                        for message_data in input_data["messages"]
                    ]
            except json.JSONDecodeError:
                self.clear()

    def clear(self):
        # clear the messages except the first which is the system prompt
        self.messages = self.messages[:1]
        self.save()

    def generate_response(self):
        messages = [m.serialize() for m in self.messages]

        response = gpt_chat_completion(messages, model=self.model)

        if "maximum context length" in response:
            return response

        message = Message("assistant", response)
        self.add_message(message.role, message.content)
        if self.history_path:
            self.save()

        if self.verbose:
            print(message.content)

        return message.content

    def get_user_messages(self):
        return [msg for msg in self.messages if msg.role == "user"]

    def summarize_conversations(
        self, num_conversations: int, summary_length: int = 200
    ) -> bool:
        if self.verbose:
            print(
                "ChatAgent.summarize_conversations: ", num_conversations, summary_length
            )

        if num_conversations <= 0:
            return False

        while num_conversations > 0:
            # Collect the first 'num_conversations' user-assistant message pairs but skip the system prompt
            conversations = self.messages[1 : (num_conversations * 2)]

            # Concatenate the user-assistant message pairs
            conversation_text = "\n".join(
                [f"{msg.role}: {msg.content}" for msg in conversations]
            )

            # Set the summarizer AGI prompt
            summarizer_prompt = "You are a Summarizer AGI, an autonomous and intelligent text summarizer."

            # Generate a summary request
            summary_request = (
                f"Please give a perfect executive summary of the salient points of the conversation in "
                f"about {summary_length} about words.:\n{conversation_text}"
            )

            # Create a message list for the summary prompt
            summary_prompt_messages = [
                {"role": "system", "content": summarizer_prompt},
                {"role": "user", "content": summary_request},
            ]

            # Submit the summary prompt to gpt_chat_completion and get the response
            summary = gpt_chat_completion(summary_prompt_messages, model=self.model)

            if self.verbose:
                print("ChatAgent.summarize_conversations: ", summary)
                print("ChatAgent.summarize: " + summary)

            # If the response contains the "maximum context length" error,
            # reduce the number of conversations and try again
            if "maximum context length" in summary:
                num_conversations -= 1
                continue

            # Check if a summary was generated
            if not summary:
                return False

            # Remove the first 'num_conversations' user-assistant message pairs from the message list
            self.messages = [self.messages[0]] + self.messages[
                (num_conversations * 2) + 1 :
            ]

            # Update the messages with the summary request and response
            self.messages.insert(
                1,
                Message(
                    "user",
                    f"Please summarize the first {num_conversations} conversations.",
                ),
            )
            self.messages.insert(2, Message("assistant", summary))

            # Save the updated chat history
            self.save()

            return True

        return False

    def __str__(self):
        return "\n".join([f"{m.role}: {m.content}" for m in self.messages])

    def __len__(self):
        return len(self.messages)

    def __getitem__(self, index):
        return self.messages[index]



import importlib.util
import os

import click
from click import Context
from rich import print

from fgn.core import dsl_parser
from fgn.core.command_context import create_context
from fgn.core.core_cmd import core_command
from fgn.utils.file_operations import create_project_dir

COMMANDS_DIRECTORY = os.path.join(os.path.dirname(__file__), "commands")


class FGNCommandsLoader(click.MultiCommand):
    def list_commands(self, ctx):
        return get_command_folders(COMMANDS_DIRECTORY)

    def get_command(self, ctx, name):
        return load_command(ctx, name, COMMANDS_DIRECTORY)


# Get command folders in the specified directory
def get_command_folders(directory):
    command_folders = [
        folder_name
        for folder_name in os.listdir(directory)
        if os.path.isdir(os.path.join(directory, folder_name))
    ]
    command_folders.sort()
    return command_folders


# Load the command with the given name from the commands_directory
def load_command(ctx, name, commands_directory):
    ctx.obj = create_context(ctx.params)

    command_script = os.path.join(commands_directory, name, f"{name}_cmd.py")

    if name == "__pycache__":
        return None

    if os.path.isfile(command_script):
        spec = importlib.util.spec_from_file_location(f"{name}_command", command_script)
        command_obj = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(command_obj)
        command = command_obj.cli
        command.obj = ctx.obj
        return command
    else:
        # Load the core command
        command_script = os.path.join(commands_directory, "core", f"core_cmd.py")
        spec = importlib.util.spec_from_file_location(f"core_command", command_script)
        command_obj = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(command_obj)
        command = command_obj.cli
        ctx.obj.text = name
        command.obj = ctx.obj
        return command


@click.group(invoke_without_command=True, cls=FGNCommandsLoader)
@click.option(
    "-m",
    "--model",
    default="gpt-4",
    help="The OpenAI model to be used for AGI response.",
)
@click.option("-i", "--input", help="Path to input file.")
@click.option("-o", "--output", help="Path to output file.")
@click.option("-io", "--in-n-out", help="Path to input and output file.")
@click.option("-pr", "--prompt", help="Prompt itself as a string.")
@click.option(
    "-v", "--verbose", is_flag=True, help="Enable verbose mode for printing output."
)
@click.option("-t", "--template", help="Path to template file.")
@click.option("-e", "--example", help="Path to example file.")
@click.option("-sp", "--system-prompt", help="Path to system prompt file.")
@click.option("-sc", "--schema", help="Path to response schema definition file.")
@click.option("--clear-history", is_flag=True, help="Clear the history file.")
@click.option(
    "-ao",
    "--auto-output",
    is_flag=True,
    help="Automatically output to file with automatic file name.",
)
@click.option(
    "-as",
    "--auto-summarize",
    type=int,
    help="Automatic summarization after the specified number of messages.",
    default=4,
)
@click.option("-p", "--paste", is_flag=True, help="Paste input from the clipboard.")
@click.option(
    "-nc", "--no-copy", is_flag=True, help="Do not copy output to the clipboard."
)
@click.option("-tk", "--tokens", help="Token replacement separated by ;")
@click.option("-ext", "--extension", help="File extension of auto output file.")
@click.option("-d", "--dsl", help="Use the FGN Domain Specific Language.")
@click.option("-a", "--append", is_flag=True, help="Append to the output file.")
@click.pass_context
def fgn(
    ctx: Context,
    model,
    input,
    output,
    prompt,
    verbose,
    no_copy,
    paste,
    example,
    template,
    clear_history,
    auto_output,
    auto_summarize,
    tokens,
    extension,
    dsl,
    system_prompt,
    schema,
    append,
    in_n_out,
    text="",
):
    fgn_ctx = create_context(ctx.params)

    create_project_dir()

    if dsl:
        dsl_parser.main(dsl, fgn_ctx)

    if verbose:
        print(fgn_ctx)


# Validate that at least one of text, input, or paste is provided
def validate_input(input, paste):
    if not input and not paste:
        raise ValueError(
            "Please provide at least one of the following: a text, an input file, or paste input."
        )


if __name__ == "__main__":
    fgn()



from dataclasses import dataclass
from typing import Dict, Optional


@dataclass
class Step:
    action: str
    description: Optional[str] = None
    run: Optional[str] = None
    command: Optional[str] = None
    options: Optional[Dict[str, str]] = None
    file_path: Optional[str] = None
    name: Optional[str] = None
    value: Optional[str] = None



# fgn/core/core_cmd.py
from time import sleep

from rich import print
from rich.columns import Columns
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.spinner import SPINNERS, Spinner
from rich.text import Text

from fgn.core.chat_agent import ChatAgent
from fgn.utils.clipboard import copy_into_clipboard, paste_into_fgn
from fgn.utils.file_operations import open_file_or_raise
from fgn.utils.openai_operations import generate_output_file


def core_command(ctx):
    print("Running core command...")

    # Use ChatAgent for handling input
    chat_agent = ChatAgent(
        system_prompt="You are a Hive-Mind Multi Agent AGI that is designed to answer questions and solve problems. "
        "You utilize emergent behavior of LLMs to come up with better answers than any single LLM. "
        "When a question has multiple answers, you will provide the best answer. You primary way to solve"
        " problems is to help humans generate code to create systems that last.",
        model=ctx.model,
        auto_clear=ctx.clear_history,
        verbose=ctx.verbose,
    )
    chat_prompt = ""
    if ctx.prompt:
        chat_prompt += ctx.prompt + " "

    if ctx.schema:
        chat_prompt += open_file_or_raise(ctx.schema) + "\n"
    if ctx.template:
        chat_prompt += open_file_or_raise(ctx.template) + "\n"
    if ctx.example:
        chat_prompt += open_file_or_raise(ctx.example) + "\n"
    if ctx.input:
        chat_prompt += open_file_or_raise(ctx.input) + "\n"
    if ctx.paste:
        chat_prompt += paste_into_fgn() + " "
    # Add text to the chat_prompt
    if ctx.text:
        chat_prompt += ctx.text + " "
    if ctx.verbose:
        print(f"Input: {chat_prompt}")

    if not chat_prompt.strip():
        raise ValueError(
            "Error: chat_prompt is empty. Please provide a prompt, input, text, or paste."
        )

    try:
        response = chat_agent.submit(chat_prompt, ctx.tokens)
    except Exception as e:
        print(f"Error: {e}")
        raise e

    if ctx.output or ctx.auto_output:
        if not ctx.output:
            ctx.output = generate_output_file(response)
        if ctx.append:
            with open(ctx.output, "a") as output_file:
                output_file.write("\n\n" + response)
        else:
            with open(ctx.output, "w") as output_file:
                output_file.write(response)
        if ctx.verbose:
            print(f"The output has been saved to {ctx.output}.")
    if not ctx.no_copy:
        copy_into_clipboard(response)

    md = Markdown(str(response))
    print(md)



import io
import os
from pathlib import Path

import click
from rich import print

from fgn.core.default_sub_cmd import default_sub_cmd

cmd_name = Path(__file__).stem.split("_")[0]


@click.command()
@click.argument("text", required=True)
@click.pass_context
def cli(ctx: click.Context, text) -> None:
    """Create a command named by the prompt."""
    ctx.obj.text = text
    new_cmd_name = ctx.obj.text
    path = Path(__file__).parent.absolute()
    new_path = path.parent.absolute() / new_cmd_name

    # Create the  ../{new_cmd_name} folder
    if not new_path.exists():
        new_path.mkdir()
    else:
        print(f"Error: {new_cmd_name} already exists.")
        exit(1)

    # Write to the config file
    with open(new_path / f"{new_cmd_name}_config.yml", "w", encoding="utf-8") as f:
        f.write(
            f"""# fgn/commands/{new_cmd_name}/{new_cmd_name}_config.yml
defaults:
  template: {new_cmd_name}_template.txt
  example: {new_cmd_name}_example.txt
  system_prompt: {new_cmd_name}_system_prompt.txt
  history_path: {new_cmd_name}_history.txt
  model: gpt-3.5-turbo
  auto_clear: false
  verbose: false
  auto_summarize: int = 4"""
        )

        # Write to the command file
        with open(new_path / f"{new_cmd_name}_cmd.py", "w", encoding="utf-8") as f:
            f.write(
                f"""# fgn/core/commands/{new_cmd_name}/{new_cmd_name}_cmd.py
from pathlib import Path
import click

from fgn.core.default_sub_cmd import default_sub_cmd

cmd_name = Path(__file__).stem.split('_')[0]


@click.command()
@click.argument('text', required=False, default='')
@click.pass_context
def cli(ctx: click.Context, text) -> None:
    \"\"\"Create {new_cmd_name}.\"\"\"
    ctx.obj.text = text
    default_sub_cmd(ctx, cmd_name)
    """
            )

        # touch all the new  files
        files = [
            f"{new_cmd_name}_example.txt",
            f"{new_cmd_name}_system_prompt.txt",
            f"{new_cmd_name}_history.txt",
            f"{new_cmd_name}_template.txt",
            "__init__.py",
        ]
        for file in files:
            os.system(f"touch {new_path}/{file}")



import logging
import os
import subprocess
import time
from typing import Dict

import yaml

from fgn.core.command_context import CommandContext
from fgn.core.core_cmd import core_command
from fgn.models.dsl.fgn_dsl_schema import FgnDslSchema
from fgn.models.dsl.step import Step
from fgn.models.dsl.task import Task

logging.basicConfig(level=logging.INFO)


def load_schema_from_yaml_file(file_path: str) -> FgnDslSchema:
    with open(file_path) as yaml_file:
        schema_dict = yaml.safe_load(yaml_file)

    tasks = []
    for task_dict in schema_dict["tasks"]:
        steps = [Step(**step_dict) for step_dict in task_dict["steps"]]
        task = Task(
            name=task_dict["name"], description=task_dict["description"], steps=steps
        )
        tasks.append(task)

    return FgnDslSchema(
        version=schema_dict["version"],
        description=schema_dict["description"],
        tasks=tasks,
    )


def execute_shell_command(
    command: str, ctx: CommandContext, variables: Dict[str, str]
) -> None:
    formatted_command = command.format(**variables)
    formatted_command = formatted_command.format(**ctx.__dict__)
    subprocess.run(
        formatted_command,
        shell=True,
        check=True,
        capture_output=True,
    )


def execute_fgn_command(
    command: str,
    options: Dict[str, str],
    ctx: CommandContext,
    variables: Dict[str, str],
) -> None:
    ctx = CommandContext(**ctx.__dict__)

    for k, v in options.items():
        v = str(v)
        setattr(ctx, k, v.format(**variables))

    for k, v in variables.items():
        v = str(v)
        setattr(ctx, k, v)

    formatted_command = command.format(**variables)
    formatted_command = formatted_command.format(**ctx.__dict__)
    formatted_options = {k: str(v).format(**variables) for k, v in options.items()}
    command_context = CommandContext(**formatted_options)
    ctx.text = formatted_command
    core_command(ctx)

    output_file_path = command_context.output
    if output_file_path:
        wait_for_file(output_file_path)


def process_step(step: Step, ctx: CommandContext, variables: Dict[str, str]) -> None:
    if step.action == "shell":
        execute_shell_command(step.run, ctx, variables)
    elif step.action == "fgn":
        execute_fgn_command(step.command, step.options, ctx, variables)
    elif step.action == "import_task":
        process_yaml_file(step.file_path, ctx, variables)
    elif step.action == "assign_variable":
        variables[step.name] = step.value
    else:
        raise ValueError(f"Unknown action '{step.action}' in step '{step.description}'")


def process_yaml_file(
    file_path: str, ctx: CommandContext, variables: Dict[str, str]
) -> None:
    schema = load_schema_from_yaml_file(file_path)
    process_schema(schema, ctx, variables)


def process_task(task: Task, ctx: CommandContext, variables: Dict[str, str]) -> None:
    logging.info(f"Processing task '{task.name}'")
    for step in task.steps:
        process_step(step, ctx, variables)


def process_schema(
    schema: FgnDslSchema, ctx: CommandContext, variables: Dict[str, str] = None
) -> None:
    if variables is None:
        variables = {}
    for task in schema.tasks:
        process_task(task, ctx, variables)


def wait_for_file(
    file_path: str, timeout: int = 10, check_interval: float = 1.0
) -> None:
    start_time = time.time()
    while not os.path.exists(file_path):
        if time.time() - start_time > timeout:
            raise TimeoutError(
                f"File '{file_path}' not found after waiting for {timeout} seconds"
            )
        time.sleep(check_interval)


def main(schema_file: str, ctx: CommandContext):
    schema = load_schema_from_yaml_file(schema_file)
    process_schema(schema, ctx)


if __name__ == "__main__":
    main("/Users/seanchatman/dev/file-generator/data/dsl/test.yaml", CommandContext())



# fgn/core/commands/chain/chain_cmd.py
from pathlib import Path

import click

from fgn.core.default_sub_cmd import default_sub_cmd

cmd_name = Path(__file__).stem.split("_")[0]


@click.command()
@click.argument("text", required=False, default="")
@click.pass_context
def cli(ctx: click.Context, text) -> None:
    """Create chain of fgn commands."""
    ctx.obj.text = f"""###
The first command must have a -o option to create the output file.
The other commands must have a -i option to read the input file and a -o option to create the output file.
Make sure that the output file is the input file for the next command.
```shell
fgn -o "<TEXT1>.md" COMMAND "<TEXT1>"
fgn -i "<TEXT1>.md" -o "<TEXT2>.md" COMMAND "<TEXT2>"
fgn -i "<TEXT2>.md" -o "<TEXT3>.md" COMMAND "<TEXT3>"
fgn -i "<TEXT3>.md" -o "<TEXT4>.md" COMMAND "<TEXT4>"
```
Provide a ```shell``` code block with the commands to run.
Use fgn to run the commands. For example:
```shell
fgn -o "buisness-plan.md" plan "Business Plan for Selling Ebooks and Classes"
fgn -i "buisness-plan.md" -o "marketing-plan.md" plan "Marketing Plan for Selling Ebooks and Classes"
fgn -i "marketing-plan.md" -o "working-backwards-press-release.md" blog "Amazon Style Working Backwards Press Release for Selling Ebooks and Classes"
fgn -i "working-backwards-press-release.md" -o "customer-success-story.md" nano "Zero to Hero Customer Success Story for Ebooks and Classes"
```

###
{text}
###
These are the chained fgn commands:
```shell
"""
    default_sub_cmd(ctx, cmd_name)



import os
import re
import uuid
from time import gmtime, sleep, strftime

import openai
from rich import print

from fgn.utils.file_operations import open_file, save_to_project_folder

openai.api_key = os.environ["OPENAI_API_KEY"]


def save_completion(prompt):
    zulu = strftime("%Y-%m-%dT%H:%M:%SZ", gmtime())
    save_to_project_folder(
        f"data/completion/completion_{uuid.uuid4()}_{zulu}.txt", prompt
    )


def save_embedding(prompt):
    zulu = strftime("%Y-%m-%dT%H:%M:%SZ", gmtime())
    save_to_project_folder(
        f"data/embeddings/embedding_{uuid.uuid4()}_{zulu}.txt", prompt
    )


def gpt3_completion(
    prompt,
    engine="text-davinci-003",
    temp=1.0,
    top_p=1.0,
    tokens=400,
    freq_pen=0.0,
    pres_pen=0.0,
    stop=None,
):
    if stop is None:
        stop = ["<<STOP>>"]
    max_retry = 3
    retry = 0
    prompt = prompt.encode(encoding="ASCII", errors="ignore").decode()
    while True:
        try:
            response = openai.Completion.create(
                engine=engine,
                prompt=prompt,
                temperature=temp,
                max_tokens=tokens,
                top_p=top_p,
                frequency_penalty=freq_pen,
                presence_penalty=pres_pen,
                stop=stop,
            )
            text = response["choices"][0]["text"].strip()
            save_completion(text)
            return text
        except Exception as oops:
            retry += 1
            if retry >= max_retry:
                return "GPT3 error: %s" % oops
            print("Error communicating with OpenAI:", oops)
            sleep(1)


def gpt_embedding(text, model="text-embedding-ada-002"):
    text = text.replace("\n", " ")
    emb = openai.Embedding.create(input=[text], model=model)["data"][0]["embedding"]
    save_embedding(emb)
    return emb


def gpt_chat_completion(
    messages, model, max_retry=5, backoff_factor=2, initial_wait=0.1
):
    """
    Sends chat inputs to OpenAI API and receives the chatbot response.

    Args:
        messages (List[Dict]): A list of messages with role and content.
        model (str, optional): The GPT model to be used.
        max_retry (int, optional): The maximum number of retries.
        backoff_factor (int, optional): The factor to exponentially increase the waiting time.
        initial_wait (float, optional): The initial waiting time between retries in seconds.

    Returns:
        str: The content of the message generated by the chatbot.
    """

    # Initialize retry attempts
    retry = 0

    # Run the loop for retry attempts
    while retry <= max_retry:
        try:
            response = openai.ChatCompletion.create(model="gpt-4", messages=messages)
            text = response["choices"][0]["message"]["content"].strip()
            save_completion(text)
            return text
        except Exception as oops:
            # If the error is due to maximum context length, return the error message so that the user can handle it
            if "maximum context length" in str(oops):
                return f"GPT error: {oops}"

            retry += 1  # Increment the retry attempts

            # If reached the maximum retry attempts, return the error message
            if retry > max_retry:
                return f"GPT error: {oops}"

            # Calculate the waiting time for exponential backoff
            wait_time = initial_wait * (backoff_factor ** (retry - 1))

            # Print the error and wait before retrying
            print(
                f"Error communicating with OpenAI (attempt {retry}/{max_retry}): {oops}"
            )
            sleep(wait_time)


def gpt4_completion(prompt, model="gpt-4"):
    completion = gpt_chat_completion([{"role": "user", "content": prompt}], model=model)
    save_completion(completion)
    return completion


def generate_filename(
    prompt, prefix="", suffix="", extension="md", max_chars=60, time=False
):
    """
    Generates a filename based on the given prompt.
    :param prompt: the prompt to generate the filename from
    :param prefix: the prefix to add to the filename
    :param suffix: the suffix to add to the filename
    :param extension: the extension to add to the filename
    :param max_chars: the maximum number of characters in the filename (default: 60)
    :param time: whether to add the current time to the filename (default: False)
    :return: the generated filename
    """
    prompt = prompt[:300]

    file_name = gpt3_completion(
        f"Carefully read the following text and generate a highly relevant and unique filename for it. The filename "
        f"should be in all lowercase letters and consist of at most {max_chars} characters, separated by underscores. "
        f"Exclude any file extensions from the filename. Pay close attention to the content "
        f"and context of the text to avoid hallucinations or mistakes. The text is: '{prompt}'\n\nfile name:",
        temp=0,
        tokens=max_chars * 10,
    )

    # Remove all non-underscore and non-alphanumeric characters, and truncate to max_chars
    file_name = re.sub(r"[^a-zA-Z0-9_]", "", file_name)
    file_name = file_name[:max_chars]

    if prefix:
        file_name = f"{prefix}_{file_name}"

    if suffix:
        file_name = f"{file_name}_{suffix}"

    if time:
        file_name = f"{file_name}_{strftime('%Y-%m-%d_%H-%M-%S', gmtime())}"

    if extension:
        file_name = f"{file_name}.{extension}"

    # print('Generated filename:', file_name)

    return file_name


def generate_output_file(prompt, extension="md", max_chars=60, time=True):
    return generate_filename(
        prompt, extension=extension, max_chars=max_chars, time=time
    )



import pytest
from click.testing import CliRunner

from fgn.cli import main


@pytest.fixture
def runner(fs):
    return CliRunner()


def test_file_writer_with_output(runner, fs):
    # Use the fs fixture provided by pytest-fs to create a fake file
    fake_file_path = "/path/to/fake_file.txt"
    fake_file_content = "Test Message"
    message = "Hello World!"
    fs.create_file(fake_file_path, contents=fake_file_content)

    # Use the runner fixture to invoke the CLI command
    result = runner.invoke(main, ["-o", fake_file_path, message])

    # Print the output and contents of the fake file for debugging
    print("result.output:", result.output)
    with open(fake_file_path) as file:
        print("fake_file_content:", file.read())

    # Verify the result
    assert result.exit_code == 0
    assert result.output.strip() == message



# fgn/core/commands/howto/howto_cmd.py
from pathlib import Path

import click

from fgn.core.default_sub_cmd import default_sub_cmd

cmd_name = Path(__file__).stem.split("_")[0]


@click.command()
@click.argument("text", required=False, default="")
@click.pass_context
def cli(ctx: click.Context, text) -> None:
    """Create a how-to article."""
    ctx.obj.text = text
    # ctx.obj.auto_output = True
    default_sub_cmd(
        ctx,
        cmd_name,
        prompt_prefix="Write a How to Article 'How to ",
        prompt_suffix="' in the format of a how-to article.\n\n# How to Article\n\n",
    )



# fgn/core/commands/prompt/prompt_cmd.py
from pathlib import Path

import click

from fgn.core.default_sub_cmd import default_sub_cmd

cmd_name = Path(__file__).stem.split("_")[0]


@click.command()
@click.argument("text", required=False, default="")
@click.pass_context
def cli(ctx: click.Context, text) -> None:
    """Create a new prompt"""
    ctx.obj.text = text
    default_sub_cmd(
        ctx,
        cmd_name,
        extract_md=True,
        prompt_suffix='\nStart with "You are a " in your response.\n',
    )



# fgn/core/commands/shell/shell_cmd.py
import platform
from os import getenv
from os.path import basename, splitext
from pathlib import Path

import click
from distro import name as distro_name

from fgn.core.default_sub_cmd import default_sub_cmd

cmd_name = Path(__file__).stem.split("_")[0]


@click.command()
@click.argument("text", required=False, default="")
@click.pass_context
def cli(ctx: click.Context, text) -> None:
    """Create a shell script."""
    operating_systems = {
        "Linux": "Linux/" + distro_name(pretty=True),
        "Windows": "Windows " + platform.release(),
        "Darwin": "Darwin/MacOS " + platform.mac_ver()[0],
    }
    current_platform = platform.system()
    os_name = operating_systems.get(current_platform, current_platform)
    shell_name = basename(getenv("SHELL", "PowerShell"))
    if os_name == "nt":
        shell_name = splitext(basename(getenv("COMSPEC", "Powershell")))[0]
    ctx.obj.text = text

    prompt_suffix = f"""###
Provide only {shell_name} commands for {os_name} without any description.
If there is a lack of details, provide most logical solution.
Ensure the output is a valid shell command.
If multiple steps required try to combine them together.
Prompt: {text}
###
Command:\n$"""

    ctx.obj.model = "gpt-3.5-turbo"
    default_sub_cmd(ctx, cmd_name, extract_md=True, prompt_suffix=prompt_suffix)



import os
import re
from pathlib import Path

import yaml

from fgn.core.command_context import CommandContext


def load_default_or_context(fgn_context: CommandContext, yaml_file):
    with open(yaml_file) as file:
        config = yaml.safe_load(file)

    defaults = config.get("defaults")

    model = (
        fgn_context.model
        if fgn_context.model
        else defaults.get("model", "gpt-3.5-turbo")
    )
    history_path = defaults.get("history_path", "history.txt")
    auto_clear = (
        fgn_context.clear_history
        if fgn_context.clear_history is not None
        else defaults.get("auto_clear", False)
    )
    verbose = (
        fgn_context.verbose
        if fgn_context.verbose is not None
        else defaults.get("verbose", False)
    )
    auto_summarize = defaults.get("auto_summarize", 4)

    system_prompt = (
        fgn_context.prompt
        if fgn_context.prompt
        else defaults.get("system_prompt", "system_prompt.txt")
    )

    example = fgn_context.example if fgn_context.example else defaults.get("example")
    template = (
        fgn_context.template if fgn_context.template else defaults.get("template")
    )
    input = fgn_context.input
    prompt = fgn_context.prompt

    return (
        model,
        history_path,
        auto_clear,
        verbose,
        auto_summarize,
        system_prompt,
        example,
        template,
        input,
        prompt,
    )


def open_file(filepath):
    filepath = get_norm_path(filepath)
    abs_path = os.path.abspath(filepath)
    with open(abs_path, encoding="utf-8") as infile:
        return infile.read()


def save_relative_to_base(filepath, content, base_path=None):
    filepath = get_norm_path(filepath)

    if base_path is None:
        base_path = os.path.dirname(os.path.abspath(__file__))

    abs_path = os.path.join(base_path, filepath)

    with open(abs_path, "w", encoding="utf-8") as outfile:
        outfile.write(content)


def open_file_or_raise(path: str) -> str:
    path = get_norm_path(path)

    if os.path.isfile(path):
        abs_path = os.path.abspath(path)
        with open(abs_path) as file:
            return file.read()
    else:
        raise FileNotFoundError("File not found")


def extract_markdown(text: str) -> str:
    markdown = re.findall(r"```(.*?)```", text, re.DOTALL)

    if len(markdown) == 0:
        return text
    else:
        markdown = [m.replace(m.split("\n")[0], "") for m in markdown]
        markdown = [m.replace("\n", "", 1) for m in markdown]
        return "\n".join(markdown)


def get_project_root() -> Path:
    return Path(__file__).parent.parent


def get_norm_path(path: str) -> str:
    if os.name == "nt":
        return os.path.normpath(path).replace("/", "\\")
    else:
        return os.path.normpath(path)


def save_to_project_folder(file_path, content):
    file = Path.home() / ".fgn" / file_path

    # Write the required directories if they don't exist
    if not os.path.exists(os.path.dirname(file)):
        os.makedirs(os.path.dirname(file))

    with open(file, "w", encoding="utf-8") as outfile:
        outfile.write(content)


def create_project_dir():
    fgn_dir = Path.home() / ".fgn"
    try:
        if not os.path.exists(fgn_dir):
            os.makedirs(fgn_dir)
    except OSError:
        print(f"Creation of the directory %s failed {fgn_dir}")
